---------------------------------------------------------------------------
--                                                                       --
--  Ast2Cfg is an ASIS application that generates a CFG out of the AST.  --
--                          http://cfg.w3x.org                           --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
--          Copyright (c) 2007, Raul Fechete, Georg Kienesberger         --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
-- This file is part of Ast2Cfg.                                         --
--                                                                       --
-- Ast2Cfg is free software; you can redistribute it and/or modify       --
-- it under the terms of the GNU General Public License as published by  --
-- the Free Software Foundation; either version 3 of the License, or     --
-- (at your option) any later version.                                   --
--                                                                       --
-- Ast2Cfg is distributed in the hope that it will be useful,            --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of        --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         --
-- GNU General Public License for more details.                          --
--                                                                       --
-- You should have received a copy of the GNU General Public License     --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>. --
--                                                                       --
---------------------------------------------------------------------------

-- A D A
with Ada.Strings.Unbounded; use Ada.Strings.Unbounded;
with Ada.Strings.Hash;

-- A S T 2 C F G
with Ast2Cfg.Utils; use Ast2Cfg.Utils;
with Ast2Cfg.Output; use Ast2Cfg.Output;
with Ast2Cfg.Id; use Ast2Cfg.Id;
with Ast2Cfg.Loop_Refinement;

package body Ast2Cfg.Post_Trans_Processing is

   package body World_Walker is

      -- Traverse a CFG Node-by-Node
      procedure Walk_Node_Rec (Curr_Node: in out Node_Object'Class) is
         Next_Nodes: Flow_Class_Ptr_List.Object;
         Next_Node:  Flow_Class_Ptr;
      begin
         Run_On_Node(Curr_Node);
         Nodes.Set_Mark(Curr_Node, 3);
         Next_Nodes := Flow_Types.Get_Succs(Flow_Types.Flow_Object(Curr_Node));
         Flow_Class_Ptr_List.Reset(Next_Nodes);
         while Flow_Class_Ptr_List.Has_Next(Next_Nodes) loop
            Flow_Class_Ptr_List.Get_Next(Next_Nodes, Next_Node);
            if (Nodes.Get_Mark(Nodes.Node_Object(Next_Node.all)) /= 3) then
               Walk_Node_Rec(Nodes.Node_Object(Next_Node.all));
            end if;
         end loop;
      end Walk_Node_Rec;

      -- Traverse the Flow Tree
      procedure Walk_Flow_Rec (Flow: in out Flow_Types.Flow_Object'Class) is
         Flows:        Flow_Class_Ptr_List.Object;
         Next_Flow:    Flow_Class_Ptr;
         Next_Node:    Node_Class_Ptr;
      begin
         if (Flow in Pkgs.Pkg_Object'Class) then
            -- This is a Package
            Run_On_Pkg(Pkgs.Pkg_Object(Flow));
         else
            -- This is a CFG
            Run_On_CFG(CFGs.CFG_Object(Flow));
            Next_Node := CFGs.Get_Head(CFGs.CFG_Object(Flow));
            if Next_Node /= null then
               Walk_Node_Rec(Next_Node.all);
            end if;
         end if;
         Flows := Get_Succs(Flow);
         Flow_Class_Ptr_List.Reset(Flows);
         while Flow_Class_Ptr_List.Has_Next(Flows) loop
            Flow_Class_Ptr_List.Get_Next(Flows, Next_Flow);
            Walk_Flow_Rec(Next_Flow.all);
         end loop;
      end Walk_Flow_Rec;

      -- Traverse all the Top Level Packages
      procedure Walk_World (World: in out Flow_World.World_Object) is
         Pkg:  Pkg_Class_Ptr             := null;
         Pkgs: Pkg_Class_Ptr_List.Object := Flow_World.Get_Pkgs(World);
      begin
         Pkg_Class_Ptr_List.Reset(Pkgs);
         while Pkg_Class_Ptr_List.Has_Next(Pkgs) loop
            Pkg_Class_Ptr_List.Get_Next(Pkgs, Pkg);
            Walk_Flow_Rec(Flow_Types.Flow_Object(Pkg.all));
         end loop;
      end Walk_World;

   end World_Walker;

   procedure Compress_Param_Trees (World: in out Flow_World.World_Object) is

      procedure Compress_Param_Tree (Curr_Node: in out Nodes.Node_Class_Ptr; Root: in Boolean) is
         Next_Nodes: Flow_Class_Ptr_List.Object;
         -- Next_Flow:  Flow_Types.Flow_Class_Ptr := null;
         -- Next_Node:  Nodes.Node_Class_Ptr      := null;
      begin

         if (Curr_Node = null) then
            return;
         end if;

         -- Info: ** Code out of Order
         --
         --       The following Code is commented out because no reccursive Processing is needed
         --
         -- Next_Nodes := Flow_Types.Get_Succs(Flow_Types.Flow_Object(Curr_Node.all));
         -- Flow_Class_Ptr_List.Reset(Next_Nodes);
         -- while Flow_Class_Ptr_List.Has_Next(Next_Nodes) loop
         --    Flow_Class_Ptr_List.Get_Next(Next_Nodes, Next_Flow);
         --    Next_Node := Nodes.Node_Object(Next_Node.all)'Access;
         --    Compress_Param_Tree(Next_Node, False);
         -- end loop;

         -- Info: ** Parameter Tree Compression
         --
         --       There are several possible Compressions (removals of uselsess Parameter Nodes);
         --       Transformation #1: Remove all empty Parameter Trees
         --       Transformation #2: Remove all Parameters of a Function if they are all empty
         --       Transformation #3: [..]

         -- Transformation #1
         Next_Nodes := Flow_Types.Get_Succs(Flow_Types.Flow_Object(Curr_Node.all));
         if ((Root = True) and
             (Flow_Class_Ptr_List.Count(Next_Nodes) = 0) and
             (Nodes.Get_Name(Curr_Node.all) = S2u(""))) then
            -- This is a Root Node with no Successors and no Name
            if (Curr_Node.all in Nodes.Param_Node_Object) then
               if (Nodes.Get_Var_Name(Nodes.Param_Node_Object(Curr_Node.all)) = "") then
                  -- This is a Parameter Node with no Variable
                  Nodes.Free(Curr_Node);
               end if;
            else
               -- This is no Parameter Node
               Nodes.Free(Curr_Node);
            end if;
         end if;

      end Compress_Param_Tree;

      procedure Process_Alias_List (Aliases: in out Flow_Types.Alias_Class_Ptr_List.Object) is
         Alias_Item: Alias_Class_Ptr := null;
         Param_Tree: Nodes.Node_Class_Ptr := null;
      begin
         Alias_Class_Ptr_List.Reset(Aliases);
         while (Alias_Class_Ptr_List.Has_Next(Aliases)) loop
            Alias_Class_Ptr_List.Get_Next(Aliases, Alias_Item);
            if Alias_Item.PTree /= null then
               Param_Tree := Nodes.Node_Object(Alias_Item.PTree.all)'Access;
               Compress_Param_Tree(Param_Tree, True);
               if Param_Tree /= null then
                  Alias_Item.PTree := Flow_Types.Flow_Object(Param_Tree.all)'Access;
               else
                  Alias_Item.PTree := null;
               end if;
            end if;
         end loop;
      end Process_Alias_List;

      procedure Process_Declarations_List (Decls: in out Flow_Types.Declaration_Ptr_List.Object);

      procedure Process_Types_List (Types: in out Flow_Types.Type_Descr_Ptr_List.Object) is
         TDescr: Type_Descr_Ptr := null;
         Param_Tree: Nodes.Node_Class_Ptr := null;
      begin
         Type_Descr_Ptr_List.Reset(Types);
         while (Type_Descr_Ptr_List.Has_Next(Types)) loop
            Type_Descr_Ptr_List.Get_Next(Types, TDescr);
            if (TDescr.Descr_Tree /= null) then
               Param_Tree := Nodes.Node_Object(TDescr.Descr_Tree.all)'Access;
               Compress_Param_Tree(Param_Tree, True);
               if Param_Tree /= null then
                  TDescr.Descr_Tree := Flow_Types.Flow_Object(Param_Tree.all)'Access;
               else
                  TDescr.Descr_Tree := null;
               end if;
            end if;
            Process_Declarations_List (TDescr.Components);
            Process_Declarations_List (TDescr.Discriminants);
         end loop;
      end Process_Types_List;

      procedure Process_Declarations_List (Decls: in out Flow_Types.Declaration_Ptr_List.Object) is
         Decl: Flow_Types.Declaration_Ptr := null;
         Param_Tree: Nodes.Node_Class_Ptr := null;
      begin
         Declaration_Ptr_List.Reset(Decls);
         while (Declaration_Ptr_List.Has_Next(Decls)) loop
            Declaration_Ptr_List.Get_Next(Decls, Decl);
            if (Decl.VType.Descr_Tree /= null) then
               Param_Tree := Nodes.Node_Object(Decl.VType.Descr_Tree.all)'Access;
               Compress_Param_Tree(Param_Tree, True);
               if Param_Tree /= null then
                  Decl.VType.Descr_Tree := Flow_Types.Flow_Object(Param_Tree.all)'Access;
               else
                  Decl.VType.Descr_Tree := null;
               end if;
            end if;
            if (Decl.Init /= null) then
               Param_Tree := Nodes.Node_Object(Decl.Init.all)'Access;
               Compress_Param_Tree(Param_Tree, True);
               if Param_Tree /= null then
                  Decl.Init := Flow_Types.Flow_Object(Param_Tree.all)'Access;
               else
                  Decl.Init := null;
               end if;
            end if;
            Process_Declarations_List(Decl.VType.Components);
         end loop;
      end Process_Declarations_List;

      procedure Process_Formal_Declarations_List (Decls: in out Flow_Types.Formal_Decl_Class_Ptr_List.Object) is
         Decl: Flow_Types.Formal_Decl_Class_Ptr := null;
         Param_Tree: Nodes.Node_Class_Ptr := null;
      begin
         Formal_Decl_Class_Ptr_List.Reset(Decls);
         while (Formal_Decl_Class_Ptr_List.Has_Next(Decls)) loop
            Formal_Decl_Class_Ptr_List.Get_Next(Decls, Decl);
            if Decl.Init /= null then
               Param_Tree := Nodes.Node_Object(Decl.Init.all)'Access;
               Compress_Param_Tree(Param_Tree, True);
               if Param_Tree /= null then
                  Decl.Init := Flow_Types.Flow_Object(Param_Tree.all)'Access;
               else
                  Decl.Init := null;
               end if;
            end if;
            if Decl.VType.Descr_Tree /= null then
               Param_Tree := Nodes.Node_Object(Decl.VType.Descr_Tree.all)'Access;
               Compress_Param_Tree(Param_Tree, True);
               if Param_Tree /= null then
                  Decl.VType.Descr_Tree := Flow_Types.Flow_Object(Param_Tree.all)'Access;
               else
                  Decl.VType.Descr_Tree := null;
               end if;
            end if;
         end loop;
      end Process_Formal_Declarations_List;

      procedure Param_Proc_Pkg (Pkg: in out Pkgs.Pkg_Object'Class) is
         Decls: Flow_Types.Declaration_Ptr_List.Object :=
           Flow_Types.Get_Decls(Flow_Types.Flow_Object(Pkg));
         FDecls: Flow_Types.Formal_Decl_Class_Ptr_List.Object :=
           Flow_Types.Get_FDecls(Flow_Types.Flow_Object(Pkg));
         Types: Flow_Types.Type_Descr_Ptr_List.Object :=
           Flow_Types.Get_Types(Flow_Types.Flow_Object(Pkg));
         Rens: Flow_Types.Alias_Class_Ptr_List.Object :=
           Flow_Types.Get_Rens(Flow_Types.Flow_Object(Pkg));
      begin
         Process_Declarations_List (Decls);
         Process_Formal_Declarations_List (FDecls);
         Process_Types_List (Types);
         Process_Alias_List (Rens);
      end Param_Proc_Pkg;

      procedure Param_Proc_CFG (CFG: in out CFGs.CFG_Object'Class) is
         Decls: Flow_Types.Declaration_Ptr_List.Object :=
           Flow_Types.Get_Decls(Flow_Types.Flow_Object(CFG));
         FDecls: Flow_Types.Formal_Decl_Class_Ptr_List.Object :=
           Flow_Types.Get_FDecls(Flow_Types.Flow_Object(CFG));
         Params: Flow_Types.Declaration_Ptr_List.Object :=
           CFGs.Get_Params(CFG);
         Types: Flow_Types.Type_Descr_Ptr_List.Object :=
           Flow_Types.Get_Types(Flow_Types.Flow_Object(CFG));
         Rens: Flow_Types.Alias_Class_Ptr_List.Object :=
           Flow_Types.Get_Rens(Flow_Types.Flow_Object(CFG));
      begin
         Process_Declarations_List (Decls);
         Process_Formal_Declarations_List (FDecls);
         Process_Declarations_List (Params);
         Process_Types_List (Types);
         Process_Alias_List (Rens);
      end Param_Proc_CFG;

      procedure Param_Proc_Node (Node: in out Nodes.Node_Object'Class) is
         Param_Tree: Nodes.Node_Class_Ptr := null;
      begin
         -- Handle the LHS Tree
         if (Node in Nodes.Assign_Node_Object'Class) then
            Param_Tree := Nodes.Get_LHS_Tree(Nodes.Assign_Node_Object(Node));
            Compress_Param_Tree(Param_Tree, True);
            Nodes.Set_LHS_Tree(Nodes.Assign_Node_Object(Node), Param_Tree);
         end if;
         -- Handle the RHS Tree
         Param_Tree := Nodes.Get_RHS_Tree(Node);
         if Param_Tree /= null then
            Compress_Param_Tree(Param_Tree, True);
            Nodes.Set_RHS_Tree(Node, Param_Tree);
         end if;
      end Param_Proc_Node;

      package Param_Walker is
         new World_Walker (Run_On_Pkg => Param_Proc_Pkg,
                           Run_On_CFG => Param_Proc_CFG,
                           Run_On_Node => Param_Proc_Node);
   begin
      Param_Walker.Walk_World(World);
   end Compress_Param_Trees;

   function Hash (X: in Nodes.Node_Class_Ptr) return Ada.Containers.Hash_Type is
   begin
      return Ada.Strings.Hash(To_String(Get_Id(X.all)));
   end Hash;

   function Equal (X: in Nodes.Node_Class_Ptr; Y: in Nodes.Node_Class_Ptr) return Boolean is
   begin
      return Get_Id(X.all) = Get_Id(Y.all);
   end Equal;

   procedure Strip_Dangling_Nodes (CFG: in out CFGs.CFG_Object'Class; Dangling_Roots: in Node_Set.Set) is
      use type Node_Set.Cursor;

      DFS_Nodes, Non_DFS_Nodes: Node_Set.Set;
      Node_Cur: Node_Set.Cursor;
      CFG_Root, Node_Ptr: Nodes.Node_Class_Ptr;
      Succs, Preds: Flow_Types.Flow_Class_Ptr_List.Object;
      Succ, Pred: Nodes.Node_Class_Ptr := null;

      procedure Dfs (Cur: in Nodes.Node_Class_Ptr; Set: in out Node_Set.Set) is
         Succs: Flow_Types.Flow_Class_Ptr_List.Object;
         Succ: Nodes.Node_Class_Ptr := null;
      begin

         -- check the node mark
         if Cur = null or else Get_Mark(Cur.all) = 1 then
            return;
         end if;

         -- set the mark of the current node
         Set_Mark(Cur.all, 1);

         -- add the current node to the node set
         Node_Set.Include(Set, Cur);

         -- recurse for all successors
         Succs := Get_Succs(Cur.all);
         Flow_Class_Ptr_List.Reset(Succs);
         while Flow_Class_Ptr_List.Has_Next(Succs) loop
            Flow_Class_Ptr_List.Get_Next(Succs, Flow_Types.Flow_Class_Ptr(Succ));
            Dfs(Succ, Set);
         end loop;
      end Dfs;
   begin

      CFG_Root := CFGs.Get_Head(CFG);

      -- Fill DFS_Nodes
      Reset_Mark(CFG_Root);
      Dfs(CFG_Root, DFS_Nodes);

      -- Fill Non_DFS_Nodes
      Node_Cur := Node_Set.First(Dangling_Roots);
      while Node_Cur /= Node_Set.No_Element loop
         Reset_Mark(CFG_Root);
         Dfs(Node_Set.Element(Node_Cur), Non_DFS_Nodes);
         Node_Set.Next(Node_Cur);
      end loop;

      -- subtract the nodes reachable through DFS
      -- to get those that are not reachable through DFS
      Node_Set.Difference(Non_DFS_Nodes, DFS_Nodes);

      Put_Line(Debug, "[Post_Trans_Processing] Removing " & Ada.Containers.Count_Type'Image(Node_Set.Length(Non_DFS_Nodes)) & " dangling nodes.");

      Put(Max, "[Post_Trans_Processing] Dangling nodes:");
      Node_Cur := Node_Set.First(Non_DFS_Nodes);
      while Node_Cur /= Node_Set.No_Element loop
         Put(Max, " " & To_String(Get_Id(Node_Set.Element(Node_Cur).all)));
         Node_Set.Next(Node_Cur);
      end loop;
      New_Line(Max);

      -- remove the dangling nodes
      Node_Cur := Node_Set.First(Non_DFS_Nodes);
      while Node_Cur /= Node_Set.No_Element loop

         Node_Ptr := Node_Set.Element(Node_Cur);
         Node_Set.Next(Node_Cur);

         -- remove predecessor links from successor
         Succs := Get_Succs(Node_Ptr.all);
         Flow_Class_Ptr_List.Reset(Succs);
         while Flow_Class_Ptr_List.Has_Next(Succs) loop
            Flow_Class_Ptr_List.Get_Next(Succs, Flow_Types.Flow_Class_Ptr(Succ));
            Rem_Pred(Succ.all, Flow_Class_Ptr(Node_Ptr));
         end loop;

         -- remove successor links from predecessor
         Preds := Get_Preds(Node_Ptr.all);
         Flow_Class_Ptr_List.Reset(Preds);
         while Flow_Class_Ptr_List.Has_Next(Preds) loop
            Flow_Class_Ptr_List.Get_Next(Preds, Flow_Types.Flow_Class_Ptr(Pred));
            Rem_Succ(Pred.all, Flow_Class_Ptr(Node_Ptr));
         end loop;

         -- remove the node
         Free(Node_Ptr);

      end loop;

   end Strip_Dangling_Nodes;

   procedure Connect_Gotos (World: in out Flow_World.World_Object) is

      procedure Run_On_Pkg (Pkg:  in out Pkgs.Pkg_Object'Class) is null;

      procedure Run_On_CFG (CFG: in out CFGs.CFG_Object'Class) is

         Sources, Targets: Nodes.Node_Class_Ptr_List.Object;

         procedure Collect (Cur: Nodes.Node_Class_Ptr) is
            Succs: Flow_Types.Flow_Class_Ptr_List.Object;
            Succ: Nodes.Node_Class_Ptr := null;
         begin

            -- check the node mark
            if Cur = null or else Get_Mark(Cur.all) = 7 then
               return;
            end if;

            -- set the mark of the current node
            Set_Mark(Cur.all, 7);

            -- if the current node is a goto add it to sources...
            if Cur.all in Nodes.Goto_Jump_Node_Object then
               Node_Class_Ptr_List.Append(Sources, Cur);
            end if;

            -- ...and if the node has a label add it to targets
            if Get_Name(Cur.all) /= "" then
               Node_Class_Ptr_List.Append(Targets, Cur);
            end if;

            -- recurse for all successors
            Succs := Get_Succs(Cur.all);
            Flow_Class_Ptr_List.Reset(Succs);
            while Flow_Class_Ptr_List.Has_Next(Succs) loop
               Flow_Class_Ptr_List.Get_Next(Succs, Flow_Types.Flow_Class_Ptr(Succ));
               Collect(Succ);
            end loop;

         end Collect;

         Src, Dst: Nodes.Node_Class_Ptr := null;
         Labels: Ustr_List.Object;
         Label: Unbounded_String;
         Succs: Flow_Types.Flow_Class_Ptr_List.Object;
         Succ: Nodes.Node_Class_Ptr := null;
         Dangling_Roots: Node_Set.Set;
      begin
         Collect(Get_Head(CFG));

         -- for every source...
         while not Node_Class_Ptr_List.Empty(Sources) loop
            Node_Class_Ptr_List.Pop(Sources, Src);

            -- ...check every target...
            Node_Class_Ptr_List.Reset(Targets);
            TARGET_LOOP:
            while Node_Class_Ptr_List.Has_Next(Targets) loop
               Node_Class_Ptr_List.Get_Next(Targets, Dst);

               -- ...get all the labels of a target...
               Labels := Tokenize(Get_Name(Dst.all), ",");

               -- ...and check each label
               while not Ustr_List.Empty(Labels) loop
                  Ustr_List.Pop(Labels, Label);

                  if Get_Target(Goto_Jump_Node_Object(Src.all)) = U2s(Label) then
                     -- we found the corresponding target so...

                     -- first, delete the edge to the existing, auxiliary successor of the goto (Src)
                     Succs := Get_Succs(Src.all);
                     Flow_Class_Ptr_List.Reset(Succs);
                     while Flow_Class_Ptr_List.Has_Next(Succs) loop
                        Flow_Class_Ptr_List.Get_Next(Succs, Flow_Types.Flow_Class_Ptr(Succ));
                        Rem_Pred(Succ.all, Flow_Class_Ptr(Src));
                        -- remember this node to remove possible dangling nodes later
                        Node_Set.Include(Dangling_Roots, Succ);
                     end loop;
                     Clear_Succs(Src.all);

                     -- connect to the target
                     Add_Succ(Src.all, Flow_Class_Ptr(Dst));
                     Add_Pred(Dst.all, Flow_Class_Ptr(Src));

                     exit TARGET_LOOP;

                  elsif Dst = Node_Class_Ptr_List.Get_Tail(Targets) and
                        Ustr_List.Empty(Labels) then
                     -- we reached the end of the list but didn't find the target
                     Put_Line(Error, "[Post_Trans_Processing] ERROR: Connect_Gotos found no target named """ &
                                     Get_Target(Goto_Jump_Node_Object(Src.all)) &
                                     """");
                  end if;
               end loop;
            end loop TARGET_LOOP;
         end loop;

         -- remove dangling nodes
         Strip_Dangling_Nodes (CFG, Dangling_Roots);
      end  Run_On_CFG;

      -- Run_On_Node is not used here, because "non-local pointer cannot
      -- point to local object". using Run_On_CFG instead.
      procedure Run_On_Node (Node: in out Nodes.Node_Object'Class) is null;

      package World_Walk is new World_Walker (Run_On_Pkg => Run_On_Pkg,
                                              Run_On_CFG => Run_On_CFG,
                                              Run_On_Node => Run_On_Node);

   begin
      World_Walk.Walk_World(World);
   end Connect_Gotos;

   procedure Refine_Loops (World: in out Flow_World.World_Object) is

      procedure Run_On_Pkg (Pkg:  in out Pkgs.Pkg_Object'Class) is null;

      procedure Run_On_CFG (CFG: in out CFGs.CFG_Object'Class) is
      begin
         Loop_Refinement.Refine(CFG);
      end Run_On_CFG;

      procedure Run_On_Node (Node: in out Nodes.Node_Object'Class) is null;

      package World_Walk is new World_Walker (Run_On_Pkg => Run_On_Pkg,
                                              Run_On_CFG => Run_On_CFG,
                                              Run_On_Node => Run_On_Node);
   begin
      World_Walk.Walk_World(World);
   end Refine_Loops;

end Ast2Cfg.Post_Trans_Processing;
